@@ -2,6 +2,9 @@
 const widgets = require('@jupyter-widgets/base');
 const reglScatterplot = require('regl-scatterplot/dist/regl-scatterplot.js');
 const pubSub = require('pub-sub-es');
+const d3Axis = require('d3-axis');
+const d3Scale = require('d3-scale');
+const d3Selection = require('d3-selection');
 const codecs = require('./codecs');
 const packageJson = require('../package.json');
 
@@ -30,6 +33,9 @@ const JupyterScatterModel = widgets.DOMWidgetModel.extend(
   },
 );
 
+const AXES_PADDING_X = 40;
+const AXES_PADDING_Y = 20;
+
 function camelToSnake(string) {
   return string.replace(/[\w]([A-Z])/g, function(m) {
     return m[0] + "_" + m[1];
@@ -62,8 +68,25 @@ function downloadBlob(blob, name) {
   document.body.removeChild(link);
 }
 
+function getScale(scaleType) {
+  if (scaleType.startsWith('log')) {
+    return d3Scale.scaleLog().base(scaleType.split('_')[1] || 10);
+  }
+
+  if (scaleType.startsWith('pow')) {
+    return d3Scale.scalePow().exponent(scaleType.split('_')[1] || 2);
+  }
+
+  return d3Scale.scaleLinear();
+}
+
 const MIN_WIDTH = 240;
 
+/**
+ * This dictionary maps between the camelCased Python property names and their
+ * JavaScript counter parts. In most cases the name is identical but they can be
+ * different. E.g., size (Python) vs pointSize (JavaScript)
+ */
 const properties = {
   backgroundColor: 'backgroundColor',
   backgroundImage: 'backgroundImage',
@@ -72,7 +95,7 @@ const properties = {
   cameraTarget: 'cameraTarget',
   cameraView: 'cameraView',
   color: 'pointColor',
-  colorActive: 'pointColorActive',
+  colorSelected: 'pointColorActive',
   colorBy: 'colorBy',
   colorHover: 'pointColorHover',
   height: 'height',
@@ -92,7 +115,7 @@ const properties = {
   sizeBy: 'sizeBy',
   connect: 'showPointConnections',
   connectionColor: 'pointConnectionColor',
-  connectionColorActive: 'pointConnectionColorActive',
+  connectionColorSelected: 'pointConnectionColorActive',
   connectionColorHover: 'pointConnectionColorHover',
   connectionColorBy: 'pointConnectionColorBy',
   connectionOpacity: 'pointConnectionOpacity',
@@ -102,8 +125,48 @@ const properties = {
   viewDownload: 'viewDownload',
   viewReset: 'viewReset',
   hovering: 'hovering',
+  axes: 'axes',
+  axesGrid: 'axesGrid',
+  xScale: 'xScale',
+  yScale: 'yScale',
 };
 
+const reglScatterplotProperty = new Set([
+  'backgroundColor',
+  'backgroundImage',
+  'cameraDistance',
+  'cameraRotation',
+  'cameraTarget',
+  'cameraView',
+  'pointColor',
+  'pointColorActive',
+  'colorBy',
+  'pointColorHover',
+  'height',
+  'lassoColor',
+  'lassoInitiator',
+  'lassoMinDelay',
+  'lassoMinDist',
+  'mouseMode',
+  'opacity',
+  'opacityBy',
+  'points',
+  'showReticle',
+  'reticleColor',
+  'selectedPoints',
+  'pointSize',
+  'sizeBy',
+  'showPointConnections',
+  'pointConnectionColor',
+  'pointConnectionColorActive',
+  'pointConnectionColorHover',
+  'pointConnectionColorBy',
+  'pointConnectionOpacity',
+  'pointConnectionOpacityBy',
+  'pointConnectionSize',
+  'pointConnectionSizeBy',
+]);
+
 // Custom View. Renders the widget model.
 const JupyterScatterView = widgets.DOMWidgetView.extend({
   render: function render() {
@@ -120,7 +183,6 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       self[propertyName] = self.model.get(camelToSnake(propertyName));
     });
 
-    this.height = this.model.get('height');
     this.width = !Number.isNaN(+this.model.get('width')) && +this.model.get('width') > 0
       ? +this.model.get('width')
       : 'auto';
@@ -139,15 +201,21 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.container.style.width = this.width === 'auto'
       ? '100%'
       : this.width + 'px';
-    this.container.style.height = this.height + 'px';
-
+    this.container.style.height = this.model.get('height') + 'px';
     this.el.appendChild(this.container);
 
+    this.canvasWrapper = document.createElement('div');
+    this.canvasWrapper.style.position = 'absolute';
+    this.canvasWrapper.style.top = '0';
+    this.canvasWrapper.style.left = '0';
+    this.canvasWrapper.style.right = '0';
+    this.canvasWrapper.style.bottom = '0';
+    this.container.appendChild(this.canvasWrapper);
+
     this.canvas = document.createElement('canvas');
     this.canvas.style.width = '100%';
     this.canvas.style.height = '100%';
-
-    this.container.appendChild(this.canvas);
+    this.canvasWrapper.appendChild(this.canvas);
 
     window.requestAnimationFrame(function init() {
       const initialOptions = {
@@ -160,7 +228,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       Object.entries(properties).forEach(function(property) {
         const pyName = property[0];
         const jsName = property[1];
-        if (self[pyName] !== null)
+        if (self[pyName] !== null && reglScatterplotProperty.has(jsName))
           initialOptions[jsName] = self[pyName];
       });
 
@@ -177,6 +245,9 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
 
       self.container.api = self.scatterplot;
 
+      if (self.model.get('axes')) self.createAxes();
+      if (self.model.get('axes_grid')) self.createAxesGrid();
+
       // Listen to events from the JavaScript world
       self.pointoverHandlerBound = self.pointoverHandler.bind(self);
       self.pointoutHandlerBound = self.pointoutHandler.bind(self);
@@ -184,6 +255,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       self.deselectHandlerBound = self.deselectHandler.bind(self);
       self.externalViewChangeHandlerBound = self.externalViewChangeHandler.bind(self);
       self.viewChangeHandlerBound = self.viewChangeHandler.bind(self);
+      self.resizeHandlerBound = self.resizeHandler.bind(self);
 
       self.scatterplot.subscribe('pointover', self.pointoverHandlerBound);
       self.scatterplot.subscribe('pointout', self.pointoutHandlerBound);
@@ -195,6 +267,14 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
         'jscatter::view', self.externalViewChangeHandlerBound
       );
 
+      if ('ResizeObserver' in window) {
+        self.canvasObserver = new ResizeObserver(self.resizeHandlerBound);
+        self.canvasObserver.observe(self.canvas);
+      } else {
+        window.addEventListener('resize', self.resizeHandlerBound);
+        window.addEventListener('orientationchange', self.resizeHandlerBound);
+      }
+
       // Listen to messages from the Python world
       Object.keys(properties).forEach(function(propertyName) {
         if (self[propertyName + 'Handler']) {
@@ -227,7 +307,180 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.model.save_changes();
   },
 
+  createAxes: function createAxes() {
+    this.axesSvg = d3Selection.select(this.container).select('svg').node()
+      ? d3Selection.select(this.container).select('svg')
+      : d3Selection.select(this.container).append('svg');
+    this.axesSvg.style('top', 0);
+    this.axesSvg.style('left', 0);
+    this.axesSvg.style('width', '100%');
+    this.axesSvg.style('height', '100%');
+    this.axesSvg.style('pointer-events', 'none');
+    this.axesSvg.style('user-select', 'none');
+    const color = this.model.get('axes_color')
+      .map(function (c) { return Math.round(c * 255); });
+    this.axesSvg.style('color', `rgba(${color[0]}, ${color[1]}, ${color[2]}, 1)`);
+
+    const { width, height } = this.container.getBoundingClientRect();
+
+    const currentXScaleRegl = this.scatterplot.get('xScale');
+    const currentYScaleRegl = this.scatterplot.get('yScale');
+
+    // Regl-Scatterplot's gl-space is always linear, hence we have to pass a
+    // linear scale to regl-scatterplot.
+    // In the future we might integrate this into regl-scatterplot directly
+    this.xScaleRegl = d3Scale.scaleLinear()
+      .domain(this.model.get('x_domain'))
+      .range([0, width - AXES_PADDING_X]);
+    // This scale is used for the D3 axis
+    this.xScaleAxis = getScale(this.model.get('x_scale'))
+      .domain(this.model.get('x_domain'))
+      .range([0, width - AXES_PADDING_X]);
+    // This scale converts between the linear, log, or power normalized data
+    // scale and the axis
+    this.xScaleRegl2Axis = getScale(this.model.get('x_scale'))
+      .domain(this.model.get('x_domain'))
+      .range(this.model.get('x_domain'));
+
+    this.yScaleRegl = d3Scale.scaleLinear()
+      .domain(this.model.get('y_domain'))
+      .range([height - AXES_PADDING_Y, 0]);
+    this.yScaleAxis = getScale(this.model.get('y_scale'))
+      .domain(this.model.get('y_domain'))
+      .range([height - AXES_PADDING_Y, 0]);
+    this.yScaleRegl2Axis = getScale(this.model.get('y_scale'))
+      .domain(this.model.get('y_domain'))
+      .range(this.model.get('y_domain'));
+
+    if (currentXScaleRegl) {
+      this.xScaleAxis.domain(
+        currentXScaleRegl.domain().map(this.xScaleRegl2Axis.invert)
+      );
+    }
+
+    if (currentYScaleRegl) {
+      this.yScaleAxis.domain(
+        currentYScaleRegl.domain().map(this.yScaleRegl2Axis.invert)
+      );
+    }
+
+    this.xAxis = d3Axis.axisBottom(this.xScaleAxis);
+    this.yAxis = d3Axis.axisRight(this.yScaleAxis);
+
+    this.xAxisContainer = this.axesSvg.select('.x-axis').node()
+      ? this.axesSvg.select('.x-axis')
+      : this.axesSvg.append('g').attr('class', 'x-axis');
+
+    this.xAxisContainer
+      .attr('transform', `translate(0, ${height - AXES_PADDING_Y})`)
+      .call(this.xAxis);
+
+    this.yAxisContainer = this.axesSvg.select('.y-axis').node()
+      ? this.axesSvg.select('.y-axis')
+      : this.axesSvg.append('g').attr('class', 'y-axis');
+
+    this.yAxisContainer
+      .attr('transform', `translate(${width - AXES_PADDING_X}, 0)`)
+      .call(this.yAxis);
+
+    this.axesSvg.selectAll('.domain').attr('opacity', 0);
+
+    this.scatterplot.set({
+      xScale: this.xScaleRegl,
+      yScale: this.yScaleRegl,
+      height: this.model.get('height') - AXES_PADDING_Y,
+    });
+
+    this.canvasWrapper.style.right = `${AXES_PADDING_X}px`;
+    this.canvasWrapper.style.bottom = `${AXES_PADDING_Y}px`;
+
+    if (this.model.get('axes_grid')) this.createAxesGrid();
+  },
+
+  removeAxes: function removeAxes() {
+    this.axesSvg.node().remove();
+    this.axesSvg = undefined;
+    this.xAxis = undefined;
+    this.yAxis = undefined;
+    this.xAxisContainer = undefined;
+    this.yAxisContainer = undefined;
+
+    this.canvasWrapper.style.top = '0';
+    this.canvasWrapper.style.left = '0';
+    this.canvasWrapper.style.right = '0';
+    this.canvasWrapper.style.bottom = '0';
+
+    this.scatterplot.set({
+      xScale: undefined,
+      yScale: undefined,
+      height: this.model.get('height'),
+    });
+  },
+
+  createAxesGrid: function createAxesGrid() {
+    const { width, height } = this.canvasWrapper.getBoundingClientRect();
+    if (this.xAxis) {
+      this.xAxis.tickSizeInner(-height);
+      this.xAxisContainer.call(this.xAxis);
+    }
+    if (this.yAxis) {
+      this.yAxis.tickSizeInner(-width);
+      this.yAxisContainer.call(this.yAxis);
+    }
+    if (this.axesSvg) {
+      this.axesSvg.selectAll('line')
+        .attr('stroke-opacity', 0.2)
+        .attr('stroke-dasharray', 2);
+    }
+  },
+
+  removeAxesGrid: function removeAxesGrid() {
+    if (this.xAxis) {
+      this.xAxis.tickSizeInner(6);
+      this.xAxisContainer.call(this.xAxis);
+    }
+    if (this.yAxis) {
+      this.yAxis.tickSizeInner(6);
+      this.yAxisContainer.call(this.yAxis);
+    }
+    if (this.axesSvg) {
+      this.axesSvg.selectAll('line')
+        .attr('stroke-opacity', null)
+        .attr('stroke-dasharray', null);
+    }
+  },
+
+  resizeHandler: function resizeHandler() {
+    if (!this.model.get('axes')) return;
+
+    const { width, height } = this.container.getBoundingClientRect();
+
+    this.xScaleAxis.range([0, width - AXES_PADDING_X]);
+    this.yScaleAxis.range([height - AXES_PADDING_Y, 0]);
+    this.xAxis.scale(this.xScaleAxis);
+    this.yAxis.scale(this.yScaleAxis);
+
+    this.xAxisContainer
+      .attr('transform', `translate(0, ${height - AXES_PADDING_Y})`)
+      .call(this.xAxis);
+    this.yAxisContainer
+      .attr('transform', `translate(${width - AXES_PADDING_X}, 0)`)
+      .call(this.yAxis);
+
+    // Render grid
+    if (this.model.get('axes_grid')) {
+      this.xAxis.tickSizeInner(-(height - AXES_PADDING_Y));
+      this.yAxis.tickSizeInner(-(width - AXES_PADDING_X));
+    }
+  },
+
   remove: function destroy() {
+    if (this.canvasObserver) {
+      this.canvasObserver.disconnect();
+    } else {
+      window.removeEventListener('resize', this.resizeHandlerBound);
+      window.removeEventListener('orientationchange', this.resizeHandlerBound);
+    }
     pubSub.globalPubSub.unsubscribe(
       'jscatter::view',
       this.externalViewChangeHandlerBound
@@ -243,11 +496,11 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
   // Helper
   colorCanvas: function colorCanvas() {
     if (Array.isArray(this.backgroundColor)) {
-      this.canvas.style.backgroundColor = 'rgb(' +
+      this.container.style.backgroundColor = 'rgb(' +
         this.backgroundColor.slice(0, 3).map(function (x) { return x * 255 }).join(',') +
         ')';
     } else {
-      this.canvas.style.backgroundColor = this.backgroundColor;
+      this.container.style.backgroundColor = this.backgroundColor;
     }
   },
 
@@ -288,15 +541,37 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
 
   viewChangeHandler: function viewChangeHandler(event) {
     const viewSync = this.model.get('view_sync');
-    if (!viewSync) return;
-    pubSub.globalPubSub.publish(
-      'jscatter::view',
-      {
-        src: this.randomStr,
-        uuid: viewSync,
-        view: event.view,
+    if (viewSync) {
+      pubSub.globalPubSub.publish(
+        'jscatter::view',
+        {
+          src: this.randomStr,
+          uuid: viewSync,
+          view: event.view,
+        }
+      );
+    }
+    if (this.model.get('axes')) {
+      this.xScaleAxis.domain(event.xScale.domain().map(this.xScaleRegl2Axis.invert));
+      this.yScaleAxis.domain(event.yScale.domain().map(this.yScaleRegl2Axis.invert));
+
+      this.xAxisContainer.call(this.xAxis.scale(this.xScaleAxis));
+      this.yAxisContainer.call(this.yAxis.scale(this.yScaleAxis));
+
+      if (this.model.get('axes_grid')) {
+        this.axesSvg.selectAll('line')
+          .attr('stroke-opacity', 0.2)
+          .attr('stroke-dasharray', 2);
       }
-    );
+    }
+  },
+
+  xScaleHandler: function xScaleHandler(newXScale) {
+    this.createAxes();
+  },
+
+  yScaleHandler: function yScaleHandler(newXScale) {
+    this.createAxes();
   },
 
   // Event handlers for Python-triggered events
@@ -339,7 +614,11 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
   },
 
   heightHandler: function heightHandler(newValue) {
-    this.withPropertyChangeHandler('height', newValue);
+    this.container.style.height = newValue + 'px';
+    this.withPropertyChangeHandler(
+      'height',
+      this.model.get('axes') ? newValue - AXES_PADDING_Y : newValue
+    );
     this.resizeHandler();
   },
 
@@ -368,7 +647,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.withPropertyChangeHandler('pointColor', newValue);
   },
 
-  colorActiveHandler: function colorActiveHandler(newValue) {
+  colorSelectedHandler: function colorSelectedHandler(newValue) {
     this.withPropertyChangeHandler('pointColorActive', newValue);
   },
 
@@ -404,7 +683,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.withPropertyChangeHandler('pointConnectionColor', newValue);
   },
 
-  connectionColorActiveHandler: function connectionColorActiveHandler(newValue) {
+  connectionColorSelectedHandler: function connectionColorSelectedHandler(newValue) {
     this.withPropertyChangeHandler('pointConnectionColorActive', newValue);
   },
 
@@ -449,6 +728,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
   },
 
   cameraRotationHandler: function cameraRotationHandler(newValue) {
+
     this.withPropertyChangeHandler('cameraRotation', newValue);
   },
 
@@ -464,6 +744,16 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.withPropertyChangeHandler('mouseMode', newValue);
   },
 
+  axesHandler: function axesHandler(newValue) {
+    if (newValue) this.createAxes();
+    else this.removeAxes();
+  },
+
+  axesGridHandler: function axesGridHandler(newValue) {
+    if (newValue) this.createAxesGrid();
+    else this.removeAxesGrid();
+  },
+
   otherOptionsHandler: function otherOptionsHandler(newOptions) {
     this.scatterplot.set(newOptions);
   },
@@ -471,9 +761,8 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
   viewDownloadHandler: function viewDownloadHandler(target) {
     if (!target) return;
 
-    const image = this.scatterplot.export();
-
     if (target === 'property') {
+      const image = this.scatterplot.export();
       this.model.set('view_data', image.data);
       this.model.set('view_shape', [image.width, image.height]);
       this.model.set('view_download', null);
